/*
Mandelbulb 3d Formula example. 
Original code by marius and jesse.
Modified by knighty. Apr 2015.
No warranty.
Compilation using mingw and msys:
-Compile with:
	gcc -c -m32 -O3 -mfpmath=387 -ffast-math uvc.c
-Verify asm code:
	objdumb -D uvc.o
-Extract machine code:
	objcopy -Obinary -j .text uvc.o uvc.bin
-Convert machine code from binary to hexadecimal:
	bin2hex uvc.bin uvc.m3f
(bin2hex is not part of msys or mingw. any other binary file editor would do the job)
-edit uvc.m3f in a text editor to add MB3D stuff. All the intermediate files are included in this archive.
*/
#include <math.h>
__attribute__((packed)) struct TIteration3Dext {
double Cw,Rold,RStopD,x,y,z,w; // use with neg indizes before C1. w is also used for 3d ifs and abox analytic DE
double Px, Py, Pz;             //      actual position, never change these!  Can be used as input.
double Cx, Cy, Cz;             //+24   these are the constants for adding. Pxyz or the julia seed.  Cw is @-56 (begin of struct)
void*  PVar;                   //+48   the actual formulas adress of constants and vars, constants at offset 0 increasing, user vars below -8 
float  SmoothItD;              //+52
double Rout;                   //+56   the square of the vector length, can be used as input
int    ItResultI;              //+64
int    maxIt;                  //+68
float  RStop;                  //+72
int    nHybrid[6];             //+76   all formulas iteration counts or single weights in interpol-hybrid
void*  fHPVar[6];              //+100  all formulas pointer to constants+vars, PVars-8=0.5, use PVar for the actual formula
void*  fHybrid[6];             //+124  the formulas adresses
int    CalcSIT;                //+148
int    DoJulia;                //+152
float  LNRStop;                //+156
int    DEoption;               //+160
float  fHln[6];                //+164  for SmoothIts
int    iRepeatFrom;            //+188
double OTrap;                  //+192
double VaryScale;              //+200  to use in vary by its
int    bFirstIt;               //+208  used also as iteration count, is set to 0 on it-start
int    bTmp;                   //+212  tmpBuf, free of use.
double Dfree1;                 //+216
double Dfree2;                 //+224
double Deriv1;                 //+232  for 4D first deriv or as full derivs
double Deriv2;                 //+240
double Deriv3;                 //+248
float  SMatrix4[4][4];         //+256  for 4d rotation, used like most other values only by the programs iteration loop procedure
};
__attribute__((packed)) struct Sconsts{
	double Pie;//just to use a constant
};
__attribute__((packed)) struct Svars{
	double cd;
	double cc;
	double cb;	// formular part
	double ca;
	double potenz;	   // mandelbulb potenz
	double iteration; // interations
	double phase;//just to use a variable
	double Dummy;//not used. What is it good for?
};



// fastcall is not quite delphi fastcall.
// first two args are ok, third is in ecx in delphi, on stack here.
void __attribute__((fastcall)) formula(
             double* x,      // [eax]
             double* y,      // [edx]
             double* arg,    // [ebp+8], points to TIteration3Dext.C1
             void* dummy     // so we end w/ ret 8 as delphi expects
             )
 {
  // Compute ptr to proper start of TIteration3Dext struct.
  struct TIteration3Dext* cfg = (struct TIteration3Dext*)(arg-7);
  // get pointer to constants.
  struct Sconsts* pconst=(struct Sconsts*) cfg->PVar;
  // get pointer to variables.
  struct Svars*   pvar=(struct Svars*) (((char*)pconst)-sizeof(struct Svars));
  // Read / write some fields as demonstration.. not a real formula ;-) ...but usable one ;D


// ;)
//http://wiki.mandelbulber.com/images/c/cc/Mandelbulber_user_guide_11-10-2010.pdf
 
   double ca,cb,cc,cd,r,a,b,xloop,yloop,zloop,ploop,iloop;
	double h, sa, sb,hh,hhh,hl,powR;
   xloop = cfg->x;
   yloop = cfg->y;
   zloop = cfg->z;
   ploop = pvar->potenz;
   iloop = pvar->iteration;
   ca = pvar->ca;
   cb = pvar->cb;
   cc = pvar->cc;
   cd = pvar->cd;



	//r
//	r = sqrty( xloop*xloop + yloop*yloop + zloop*zloop );
	h = xloop*xloop;
	hh = yloop*yloop;
	hhh = zloop*zloop;
	r = h+hh;
	h = r+hhh;
	r = sqrt(h);

	//a
//	a = atan2y( yloop, xloop );
	a = atan2 ( yloop, xloop );

	//b
//	b = -1*atan2y( zloop, sqrty( xloop*xloop + yloop*yloop ) );
	h = xloop*xloop;
	hh = yloop*yloop;
	hhh = h + hh;
	h = sqrt(hhh);
	hh = atan2(zloop, h);
	hhh = -1*hh;
	b = hhh;


	//pow
	h = r-1;
	hh = h*h;

	hhh = h-hh;
	hl = h*h;
	hh = hl*h;
	powR = hhh+hh;
	hh = powR*ploop;

	powR = 1+hh;
	h = hh*hh;
	hhh = powR +h;
	hl = h*hh;
	powR = hl +hhh;



	//x
//	xloop = powy( r, ploop ) * cosy( ploop * b ) * cosy( ploop * a ) + ca;
	h = powR;
	hh = ploop*b;
	hhh = cos(hh);
	xloop = h*hhh;
	h = ploop*a;
	hh = cos(h);
	hhh = xloop * hh; 
	xloop = hhh + ca;


	//y
//	yloop = powy( r, ploop ) * cosy( ploop * b ) * siny( ploop * a ) + cb;
	h = powR;
	hh = ploop*b;
	hhh = cos(hh);
	yloop = h*hhh;
	h = ploop*a;
	hh = sin(h);
	hhh = yloop * hh;
	yloop = hhh +cb;



	//z
//	zloop = powy( r, ploop ) * siny( ploop * b ) + cc;
	h = powR;
	hh = ploop * b;
	hhh = sin (hh);
	hh = h * hhh;
	zloop = hh + cc;

   cfg->x = xloop * cd;
   cfg->y = yloop * cd;
   cfg->z = zloop * cd;


}


